/*
 * Dead Reckoning program for Little Andy Robot
 * Version: 0.1.4
 * Author: Douglas E Paradis
 * Copyright 2012 Douglas E. Paradis, all rights reserved
 *
 * Use of this program is allowed for non-commerical purposes
 * as long as copyright notice is retained.
 *
 * Description: This program uses a delta theta PID controller
 * to allow the robot to drive in a straight line. Encoder counts
 * are used to determine the distance the robot has travelled in 
 * a straight line.
 *
 * Turns are made by turning the robot a given number of encoder 
 * counts with the motors driving in opposite directions. The angle
 * of the turns is based on the number of counts specified.
 *
 * Travel in a straight line and turns are executed as different states.
 * Paths consisting of a number of straight portions and turns can be
 * constructed by changing the travel arrays.
 *
 * This program is not a navigation program. It does not track position
 * in terms of X, Y, and theta from the starting position.
 *
 * This program was used by the Little Andy robot in the Dallas Personal
 * Robotics Group's (www.dprg.org) Roborama2011b competition. The robot
 * was featured in the 2012 March/April issue of Robot Magazine.   
 * 
 * Hardware:  Pololu Orangutan SVP-1284 robot controller
 *            Pololu 75:1 Metal Gearmotors with 48 CPR Encoders
 *            Pololu 70mm wheels and hubs
 *            Pololu motor brackets
 *            dougparadis.com Little Andy robot chassis kit
 *
 * Libraries: Pololu AVR C/C++ Library
 *
 * Notes:  
 *  motor 1 (left from rear) is connected to encoder cd
 *  motor 2 (right from rear) is connected to encoder ab 
 *
 *  Robot speed is set by adjusting the user trim pot on the 
 *  Orangutan SVP board.
 *
 */


#include <pololu/orangutan.h>
#include <avr/interrupt.h>




#define ON 1
#define OFF 0
#define ARR_LENGTH(x) (sizeof(x)/sizeof(*(x)))

/************** Encoder Calibrations ********************/
// Results of 90 degree turn with 3 foot straight 
// and UMNMark calibration for robot 
// note: individual robots will have different values
#define ENC_CNT_PER_DEG_CW      25        //25 // 2250 cnts per 90 deg
#define ENC_CNT_PER_DEG_CCW     26        //26 // 2340 cnts per 90 deg
#define ENC_CNT_PER_INCH_FWD   416        //415.806 // circum = 8.657874 inches, 3600 cnts /rev  
// A positive adjustment adds to the cnt, a negative adjustment substracts from the count
// You can't have a zero adj, if you need zero, make it a very large number
#define ADJ_ONE_PER_X_DEG_CW    32000        // Adjustment for error caused by integer rounding
#define ADJ_ONE_PER_X_DEG_CCW   32000           //  " 
#define ADJ_ONE_PER_X_INCH_FWD  -5           //  "
/********************************************************/


// function prototypes
void initize_robot(void);
void display_LCD(void);
int calc_error(int m0_int_cnt, int m1_int_cnt);
void pid_loop(void);
void reset(void);
void make_turn(int turn_index);
void go_straight(int segment_index);


int mtr1_total_cnt, mtr2_total_cnt;
int mtr1_interval_cnt, mtr2_interval_cnt;
int mtr1_past_cnt = 0, mtr2_past_cnt = 0;
int top_speed = 0;
int error, last_error, cum_error;
int segment_index = 0;

int turn_flg = 0;

/************** Travel Array ****************************/
// The robot's path is defined by an array of values that
// describe length of straight travel in inches and turns in
// degrees. For turns, CW turns are positive and CCW turns are
// negative. Example paths are commented out below.

// travel_arr format is {straight in inches, turn in degees, ....,straight in inches, turn in degees}. 

 int travel_arr[] = {60,0};                         // go straight 5 ft
// int  travel_arr[] = {120,0};                         // go straight 10 ft
// int travel_arr[] = {36,90,36,0};                // go 3 ft, turn 90 deg CW, and go 3 more ft
// int travel_arr[] = {36,-90,36,0};                // go 3 ft, turn 90 deg CCW, and go 3 more ft
// int travel_arr[] = {36,-540,36,0,0,180};          // go 3 ft, turn 1.5 circles CCW, and come backo 3 ft and turn 180 deg
// int travel_arr[] = {24,90,24,-90,24,-90,24,-90,24,-90,24,90,24,90,24,90}; // 2ft block figure eight
// int travel_arr[] = {48,90,48,-90,48,-90,48,-90,48,-90,48,90,48,90,48,90}; // 4ft block figure eight
// int travel_arr[] = {36,-90,36,-90,36,-90,36,-90};   // 3 foot square CCW
// int travel_arr[] = {36,90,36,90,36,90,36,90};   // 3 foot square CW
// int travel_arr[] = {36,90,36,90,36,90,36,180,36,-90,36,-90,36,-90,36,-180};  // 3 ft sq CW followed by 3 ft sq CCW


int path_seg_cnt = ARR_LENGTH(travel_arr);
/********************************************************/


/************** PID Parameters **************************/
// note: individual robots will have different values
int Kp = 6;      
int Kd = 4;   
int Ki = 1;    
/********************************************************/

int main()
{
  svp_set_mode(SVP_MODE_ENCODERS);


  initize_robot();


  int turn_index = 1;  // first turn cmd index in travel_arr
  

  while(1)
  {

     go_straight(segment_index);   
     
	 make_turn(turn_index);
	 
	 segment_index += 2;
	 turn_index = segment_index + 1;


	 // stop at end of path
     if (segment_index > (path_seg_cnt - 1))    // meed to subtract 1, array index starts with 0
	 {
	     set_motors (0,0);

		 clear(); 
		 lcd_goto_xy(0,0);
	     print("Finished - Reset");
		 lcd_goto_xy(0,1);
         print("to run again");

    
         // cause robot to stop - do not hit TOP_BUTTON, reset instead. 
         unsigned char button = wait_for_button_press(TOP_BUTTON);
         wait_for_button_release(button);  // wait 


     }
 

  }


}



void display_LCD(void) { 

    clear();
    
	lcd_goto_xy(0,0);
	print("L");
    print_long(mtr1_interval_cnt);  // mtr 1 encoder counts for interval
    print(" ");

    lcd_goto_xy(8,0);
	print("R");
    print_long(mtr2_interval_cnt);  // mtr 2 encoder counts for interval
    print(" ");

    lcd_goto_xy(0,1);
    print("E");
    print_long(error);               // error for interval
    print(" ");

    lcd_goto_xy(8,1);
    print("CE");
    print_long(cum_error);           // cummulative error 
    print(" ");
	
    delay_ms(5);
}


int calc_error(int m1_int_cnt, int m2_int_cnt)
{
   int err = 0;

   /* Error is equal to ref minus feedback. Since we are trying to make delta theta equal zero
      reference is implied to be zero and any difference in encoder cnts between motors is error. 
      NOTE: 
           if m2 (right motor) is faster (ie. more neg) the error will be negative
           if m2 is slower (ie. less neg) the error will be positive
   */

   err = (m1_int_cnt + m2_int_cnt);   // left (m1) has + cnts, right (m2) has - cnts

   return err;
}

void pid_loop (void) 
{

	int errort=0;

	errort = (Kp * (error) + Kd * (last_error - error) + Ki * (cum_error + error)); 

	
	last_error = error;
	cum_error += error;

	// limit cummulative error (prevent integrator windup)
	if (cum_error >= 200)   
	    cum_error = 200;
    else if (cum_error <= -200)
	    cum_error = -200;

	if(errort > top_speed)              // limit max error to top speed value
		errort = top_speed;
	else if(errort < -top_speed)
		errort = -top_speed;

    lcd_goto_xy(8,1);
    print_long(errort);
    print(" ");

	if(errort>0)                    // robot needs to turn left   (>)
	{
	    set_motors(top_speed, top_speed + errort);  // speed up right motor
		green_led(ON);                     
		red_led(OFF);                      
	}
	else if(errort<0)               // robot needs to turn right   (<)
	{
	    set_motors(top_speed - errort, top_speed);   //  slow down right motor 
		red_led(ON);                        
		green_led(OFF); 
	}

	else                            // go straight
	{
	    set_motors(top_speed, top_speed);  // set speeds of motors 1 and 2
		red_led(OFF);
	    green_led(OFF);
	}


}




void reset(void)
{
	/*  This function flashes the red and green LEDs a few times.
	    It is useful for troubleshooting to determine if the controller 
		is constantly resetting.
	 */
	 
	    
    // turn status LEDs off
	green_led(ON);
	red_led(ON);
	delay_ms(300);


    // Flash status LEDSs a few times
    for (char j = 0; j < 3; j++)  // even # LEDs off
	{
	green_led(j % 2);
	red_led(j % 2);
	delay_ms(500);
    }

	// leave status LEDs off
	green_led(OFF);
	red_led(OFF);

}


void initize_robot(void) 
{

  set_motors(0,0);  // stop robot

  reset();
   
  clear();                         // clear lcd
  print("Waiting");
    
  // wait for the middle button to be pressed
  // store the value of the pressed button in the variable 'button'
  unsigned char button = wait_for_button_press(MIDDLE_BUTTON);
  clear();

  wait_for_button_release(button);  // wait for middle button to be released 
  print("released");      // display that the button was released
  delay_ms(1000);
  clear();

  delay_ms(1500);


  // ---------- set top_speed ----------------
  // note that the following line could also be accomplished with:
  // int pot = analogRead(7);
  int pot = read_trimpot();    // determine the trimpot position
  delay_ms(50);
  top_speed = pot/2-256;  // turn pot reading into number between -256 and 255
  if(top_speed == -256)
      top_speed = -255; // 256 is out of range

}



void make_turn (int turn_index)
{
   set_motors(0,0);  // stop robot

   delay_ms(1500);   // wait 1.5 seconds
   
   mtr1_total_cnt = svp_get_counts_and_reset_cd();      //reset counter for encoder cd
   mtr2_total_cnt = svp_get_counts_and_reset_ab();      //reset counter for encoder ab

   mtr1_total_cnt = 0;
   mtr2_total_cnt = 0;

   if (travel_arr[turn_index] >= 1)    // cw turn
   {
       long turn_enc_cnt = travel_arr[turn_index] * ENC_CNT_PER_DEG_CW + travel_arr[turn_index] / ADJ_ONE_PER_X_DEG_CW;        
		
		//-------------
		
	   int num_resets_needed = turn_enc_cnt / 28800;   // divides dist_to_go into ~5 ft segments (8 wheel revolutions)

       int rem_enc_cnts = turn_enc_cnt % 28800;   // leftover enc cnts


       set_motors(top_speed/2, -top_speed/2);  // half speed cw

       while (num_resets_needed > 0)
       {
             mtr1_total_cnt = svp_get_counts_cd();                 // pos num  (right motor - mtr1)
	         mtr2_total_cnt = svp_get_counts_ab();                 // neg num  (left motor - mtr2)


             // for long distances you have to reset the encoders to prevent overrunning the timer
             if (svp_get_counts_cd() >= 28800) // 8 wheel revolutions (right motor)
	         {
                   num_resets_needed--;
	  
	               mtr1_total_cnt = svp_get_counts_and_reset_cd();      //reset counter for encoder cd
                   mtr2_total_cnt = svp_get_counts_and_reset_ab();      //reset counter for encoder ab	

	         }
	
	   } 
		
	   while (rem_enc_cnts > 0)
       {
            mtr1_total_cnt = svp_get_counts_cd();                 // pos num  (right motor - mtr1)
	        mtr2_total_cnt = svp_get_counts_ab();                 // neg num  (left motor - mtr2)


		    if ((mtr1_total_cnt >= rem_enc_cnts) && (mtr2_total_cnt <= rem_enc_cnts))
	        {
			    set_motors(0,-top_speed/4);  // stop left motor
	            //rem_enc_cnts = 0;
            }
			
			if ((mtr1_total_cnt <= rem_enc_cnts) && (mtr2_total_cnt >= rem_enc_cnts))
	        {
			    set_motors(top_speed/4,0);  // stop right motor 
	            //rem_enc_cnts = 0;
            }
			
			if ((mtr1_total_cnt >= rem_enc_cnts) && (mtr2_total_cnt >= rem_enc_cnts))
	        {
			    set_motors(0,0);  // stop both motors
	            rem_enc_cnts = 0;
            }
		  
		  	 
       }
       
   }	   	
		      

   if (travel_arr[turn_index] <= -1)    // ccw turn
   {
       int turn_enc_cnt = -(travel_arr[turn_index] * ENC_CNT_PER_DEG_CCW + travel_arr[turn_index] / ADJ_ONE_PER_X_DEG_CCW);

	    
	   set_motors(-top_speed/2, top_speed/2);  // half speed ccw
		
	   int num_resets_needed = turn_enc_cnt / 28800;   // divides dist_to_go into ~5 ft segments (8 wheel revolutions)

       int rem_enc_cnts = turn_enc_cnt % 28800;   // leftover enc cnts


       set_motors(-top_speed/2, top_speed/2);  // half speed cw

       while (num_resets_needed > 0)
       {
             mtr1_total_cnt = svp_get_counts_cd();                 // pos num  (right motor - mtr1)
	         mtr2_total_cnt = svp_get_counts_ab();                 // neg num  (left motor - mtr2)

             // for long distances you have to reset the encoders to prevent overrunning the timer
             if (svp_get_counts_cd() >= 28800) // 8 wheel revolutions (right motor)
	         {
                   num_resets_needed--;
	  
	               mtr1_total_cnt = svp_get_counts_and_reset_cd();      //reset counter for encoder cd
                   mtr2_total_cnt = svp_get_counts_and_reset_ab();      //reset counter for encoder ab	

	         }
	
	   } 
		
	   while (rem_enc_cnts > 0)
       {
            mtr1_total_cnt = svp_get_counts_cd();                 // pos num  (right motor - mtr1)
	        mtr2_total_cnt = svp_get_counts_ab();                 // neg num  (left motor - mtr2)


		    if ((mtr1_total_cnt <= -rem_enc_cnts) && (mtr2_total_cnt >= -rem_enc_cnts))
	        {
			    set_motors(0,top_speed/4);  // stop left motor
	            //rem_enc_cnts = 0;
            }
			
			if ((mtr1_total_cnt >= -rem_enc_cnts) && (mtr2_total_cnt <= -rem_enc_cnts))
	        {
			    set_motors(-top_speed/4,0);  // stop right motor 
	            //rem_enc_cnts = 0;
            }
			
			if ((mtr1_total_cnt <= -rem_enc_cnts) && (mtr2_total_cnt <= -rem_enc_cnts))
	        {
			    set_motors(0,0);  // stop both motors
	            rem_enc_cnts = 0;
            }


       }
       


    }



    //set_motors(0,0);  // stop robot

    delay_ms(1500);   // wait 1.5 seconds

    clear();          // clear the LCD


}


void go_straight (int segment_index)
{

   set_motors(0,0);  // stop robot

   mtr1_total_cnt = svp_get_counts_and_reset_cd();      //reset counter for encoder cd
   mtr2_total_cnt = svp_get_counts_and_reset_ab();      //reset counter for encoder ab

   mtr1_total_cnt = 0;
   mtr2_total_cnt = 0;
   
   long dist_to_go = travel_arr[segment_index] * ENC_CNT_PER_INCH_FWD + travel_arr[segment_index] / ADJ_ONE_PER_X_INCH_FWD;


   // divide dist_to_go into ~5 ft segments (8 wheel revolutions) to prevent timer overflow
   int num_resets_needed = dist_to_go / 28800;   // divides dist_to_go into ~5 ft segments (8 wheel revolutions)

   int rem_enc_cnts = dist_to_go % 28800;   // leftover enc cnts

   set_motors(top_speed, top_speed);  // start up robot

   while (num_resets_needed > 0)
   {
      mtr1_total_cnt = svp_get_counts_cd();                 // pos num  (right motor - mtr1)
	  mtr2_total_cnt = svp_get_counts_ab();                 // neg num  (left motor - mtr2)

	  mtr1_interval_cnt = mtr1_total_cnt - mtr1_past_cnt;
	  mtr2_interval_cnt = mtr2_total_cnt - mtr2_past_cnt;

	  mtr1_past_cnt = mtr1_total_cnt;
	  mtr2_past_cnt = mtr2_total_cnt;

	  error = calc_error(mtr1_interval_cnt,mtr2_interval_cnt);

      pid_loop();

	  display_LCD();

      // for long distances you have to reset the encoders to prevent overrunning the timer
      if (svp_get_counts_cd() >= 28800) // 8 wheel revolutions (right motor)
	  {
         num_resets_needed--;
	  
	     mtr1_total_cnt = svp_get_counts_and_reset_cd();      //reset counter for encoder cd
         mtr2_total_cnt = svp_get_counts_and_reset_ab();      //reset counter for encoder ab	
	
	     mtr1_past_cnt = 0;
	     mtr2_past_cnt = 0;

	     cum_error = 0;
         last_error = 0;
	     error = 0;
	   }
	
	} 
    
	
   while (rem_enc_cnts > 0)
   {
      mtr1_total_cnt = svp_get_counts_cd();                 // pos num  (right motor - mtr1)
	  mtr2_total_cnt = svp_get_counts_ab();                 // neg num  (left motor - mtr2)

	  mtr1_interval_cnt = mtr1_total_cnt - mtr1_past_cnt;
	  mtr2_interval_cnt = (mtr2_total_cnt - mtr2_past_cnt);

	  mtr1_past_cnt = mtr1_total_cnt;
	  mtr2_past_cnt = mtr2_total_cnt;

	  error = calc_error(mtr1_interval_cnt,mtr2_interval_cnt);

      pid_loop();

	  display_LCD();

	  if ((mtr1_total_cnt >= rem_enc_cnts) || (mtr2_total_cnt <= -rem_enc_cnts))
	  {
	     rem_enc_cnts = 0;
      } 
   }
       

   cum_error = 0;
   last_error = 0;
   error = 0;
      
 

}


